(function() {"use strict";var __module = CC_EDITOR ? module : {exports:{}};var __filename = 'preview-scripts/assets/Script/ListView.js';var __require = CC_EDITOR ? function (request) {return cc.require(request, require);} : function (request) {return cc.require(request, __filename);};function __define (exports, require, module) {"use strict";
cc._RF.push(module, 'fc6a7WPx2lPaZHdmN7Pxe4Y', 'ListView', __filename);
// Script/ListView.ts

Object.defineProperty(exports, "__esModule", { value: true });
var _a = cc._decorator, ccclass = _a.ccclass, property = _a.property;
/**
 * 通用 ListView 组件.
 * 能够显示垂直/横向ListView. 具体用法见Demo
 */
var ListView = /** @class */ (function (_super) {
    __extends(ListView, _super);
    function ListView() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.itemTemplate = null;
        _this.spacing = 1;
        // 比可见元素多缓存3个, 缓存越多,快速滑动越流畅,但同时初始化越慢.
        _this.spawnCount = 2;
        _this.scrollView = null;
        _this.content = null;
        _this.adapter = null;
        _this._items = new cc.NodePool();
        // 记录当前填充在树上的索引. 用来快速查找哪些位置缺少item了.
        _this._filledIds = {};
        _this.horizontal = false;
        // 初始时即计算item的高度.因为布局时要用到.
        _this._itemHeight = 1;
        _this._itemWidth = 1;
        _this._itemsVisible = 1;
        _this.lastStartIndex = -1;
        _this.scrollTopNotifyed = false;
        _this.scrollBottomNotifyed = false;
        _this.pullDownCallback = null;
        _this.pullUpCallback = null;
        return _this;
    }
    ListView.prototype.onLoad = function () {
        if (this.scrollView) {
            this.content = this.scrollView.content;
            this.horizontal = this.scrollView.horizontal;
            if (this.horizontal) {
                this.scrollView.vertical = false;
                this.content.anchorX = 0;
                this.content.anchorY = 0.5;
                this.content.x = 0 - this.content.getParent().width * this.content.getParent().anchorX;
            }
            else {
                this.scrollView.vertical = true;
                this.content.anchorX = 0.5;
                this.content.anchorY = 1;
                this.content.y = this.content.getParent().height * this.content.getParent().anchorY;
            }
        }
        else {
            console.error("ListView need a scrollView for showing.");
        }
        var itemOne = this._items.get() || cc.instantiate(this.itemTemplate);
        this._items.put(itemOne);
        this._itemHeight = itemOne.height || 10;
        this._itemWidth = itemOne.width || 10;
        if (this.horizontal) {
            this._itemsVisible = Math.ceil(this.content.getParent().width / this._itemWidth);
        }
        else {
            this._itemsVisible = Math.ceil(this.content.getParent().height / this._itemHeight);
        }
        console.log("可见区域的item数量为:", this._itemsVisible);
        this.adjustEvent();
    };
    ListView.prototype.setAdapter = function (adapter) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                this.adapter = adapter;
                if (this.adapter == null) {
                    console.warn("adapter 为空.");
                    return [2 /*return*/];
                }
                if (this.itemTemplate == null) {
                    console.error("Listview 未设置待显示的Item模板.");
                    return [2 /*return*/];
                }
                this.notifyUpdate();
                return [2 /*return*/];
            });
        });
    };
    ListView.prototype.getItemIndex = function (height) {
        return Math.floor(Math.abs(height / ((this._itemHeight + this.spacing))));
    };
    ListView.prototype.getPositionInView = function (item) {
        var worldPos = item.getParent().convertToWorldSpaceAR(item.position);
        var viewPos = this.scrollView.node.convertToNodeSpaceAR(worldPos);
        return viewPos;
    };
    // 数据变更了需要进行更新UI显示, 可只更新某一条.
    ListView.prototype.notifyUpdate = function (updateIndex) {
        var _this = this;
        if (this.adapter == null) {
            return;
        }
        if (updateIndex && updateIndex.length > 0) {
            updateIndex.forEach(function (i) {
                if (_this._filledIds.hasOwnProperty(i)) {
                    delete _this._filledIds[i];
                }
            });
        }
        else {
            Object.keys(this._filledIds).forEach(function (key) {
                delete _this._filledIds[key];
            });
        }
        this.recycleAll();
        this.lastStartIndex = -1;
        if (this.horizontal) {
            this.content.width = this.adapter.getCount() * (this._itemWidth + this.spacing) + this.spacing;
        }
        else {
            this.content.height = this.adapter.getCount() * (this._itemHeight + this.spacing) + this.spacing; // get total content height
        }
        this.scrollView.scrollToTop();
    };
    ListView.prototype.scrollToTop = function (anim) {
        if (anim === void 0) { anim = false; }
        this.scrollView.scrollToTop(anim ? 1 : 0);
    };
    ListView.prototype.scrollToBottom = function (anim) {
        if (anim === void 0) { anim = false; }
        this.scrollView.scrollToBottom(anim ? 1 : 0);
    };
    ListView.prototype.scrollToLeft = function (anim) {
        if (anim === void 0) { anim = false; }
        this.scrollView.scrollToLeft(anim ? 1 : 0);
    };
    ListView.prototype.scrollToRight = function (anim) {
        if (anim === void 0) { anim = false; }
        this.scrollView.scrollToRight(anim ? 1 : 0);
    };
    // 下拉事件.
    ListView.prototype.pullDown = function (callback, this$) {
        this.pullDownCallback = callback.bind(this$);
    };
    // 上拉事件.
    ListView.prototype.pullUp = function (callback, this$) {
        this.pullUpCallback = callback.bind(this$);
    };
    ListView.prototype.update = function (dt) {
        var startIndex = this.checkNeedUpdate();
        if (startIndex >= 0) {
            this.updateView(startIndex);
        }
    };
    // 向某位置添加一个item.
    ListView.prototype._layoutVertical = function (child, posIndex) {
        this.content.addChild(child);
        // 增加一个tag 属性用来存储child的位置索引.
        child["_tag"] = posIndex;
        this._filledIds[posIndex] = posIndex;
        child.setPosition(0, -child.height * (0.5 + posIndex) - this.spacing * (posIndex + 1));
    };
    // 向某位置添加一个item.
    ListView.prototype._layoutHorizontal = function (child, posIndex) {
        this.content.addChild(child);
        // 增加一个tag 属性用来存储child的位置索引.
        child["_tag"] = posIndex;
        this._filledIds[posIndex] = posIndex;
        child.setPosition(child.width * (child.anchorX + posIndex) + this.spacing * posIndex, 0);
    };
    // 获取可回收item
    ListView.prototype.getRecycleItems = function (beginIndex, endIndex) {
        var _this = this;
        var children = this.content.children;
        var recycles = [];
        children.forEach(function (item) {
            if (item["_tag"] < beginIndex || item["_tag"] > endIndex) {
                recycles.push(item);
                delete _this._filledIds[item["_tag"]];
            }
        });
        return recycles;
    };
    ListView.prototype.recycleAll = function () {
        var _this = this;
        var children = this.content.children;
        this.content.removeAllChildren();
        children.forEach(function (item) {
            _this._items.put(item);
        });
    };
    // 填充View.
    ListView.prototype.updateView = function (startIndex) {
        var _this = this;
        var itemStartIndex = startIndex;
        // 比实际元素多3个.
        var itemEndIndex = itemStartIndex + this._itemsVisible + (this.spawnCount || 2);
        var totalCount = this.adapter.getCount();
        if (itemStartIndex >= totalCount) {
            return;
        }
        if (itemEndIndex > totalCount) {
            itemEndIndex = totalCount;
            if (itemStartIndex > 0 && (!this.scrollBottomNotifyed)) {
                this.notifyScrollToBottom();
                this.scrollBottomNotifyed = true;
            }
        }
        else {
            this.scrollBottomNotifyed = false;
        }
        // 回收需要回收的元素位置.向上少收一个.向下少收2两.
        var recyles = this.getRecycleItems(itemStartIndex - (this.spawnCount || 2), itemEndIndex);
        recyles.forEach(function (item) {
            _this._items.put(item);
        });
        // 查找需要更新的元素位置.
        var updates = this.findUpdateIndex(itemStartIndex, itemEndIndex);
        // 更新相应位置.
        for (var _i = 0, updates_1 = updates; _i < updates_1.length; _i++) {
            var index = updates_1[_i];
            var child = this.adapter._getView(this._items.get() || cc.instantiate(this.itemTemplate), index);
            this.horizontal ?
                this._layoutHorizontal(child, index) :
                this._layoutVertical(child, index);
        }
    };
    // 检测是否需要更新UI.
    ListView.prototype.checkNeedUpdate = function () {
        if (this.adapter == null) {
            return -1;
        }
        var scroll = this.horizontal ? (-this.content.x - this.content.getParent().width * this.content.getParent().anchorX)
            : (this.content.y - this.content.getParent().height * this.content.getParent().anchorY);
        var itemStartIndex = Math.floor(scroll / ((this.horizontal ? this._itemWidth : this._itemHeight) + this.spacing));
        if (itemStartIndex < 0 && !this.scrollTopNotifyed) {
            this.notifyScrollToTop();
            this.scrollTopNotifyed = true;
            return itemStartIndex;
        }
        // 防止重复触发topNotify.仅当首item不可见后才能再次触发
        if (itemStartIndex > 0) {
            this.scrollTopNotifyed = false;
        }
        if (this.lastStartIndex != itemStartIndex) {
            this.lastStartIndex = itemStartIndex;
            return itemStartIndex;
        }
        return -1;
    };
    // 查找需要补充的元素索引.
    ListView.prototype.findUpdateIndex = function (itemStartIndex, itemEndIndex) {
        var d = [];
        for (var i = itemStartIndex; i < itemEndIndex; i++) {
            if (this._filledIds.hasOwnProperty(i)) {
                continue;
            }
            d.push(i);
        }
        return d;
    };
    ListView.prototype.notifyScrollToTop = function () {
        if (!this.adapter || this.adapter.getCount() <= 0) {
            return;
        }
        if (this.pullDownCallback) {
            this.pullDownCallback();
        }
    };
    ListView.prototype.notifyScrollToBottom = function () {
        if (!this.adapter || this.adapter.getCount() <= 0) {
            return;
        }
        if (this.pullUpCallback) {
            this.pullUpCallback();
        }
    };
    ListView.prototype.adjustEvent = function () {
        var _this = this;
        this.content.on(this.isMobile() ? cc.Node.EventType.TOUCH_END : cc.Node.EventType.MOUSE_UP, function () {
            _this.scrollTopNotifyed = false;
            _this.scrollBottomNotifyed = false;
        }, this);
        this.content.on(this.isMobile() ? cc.Node.EventType.TOUCH_CANCEL : cc.Node.EventType.MOUSE_LEAVE, function () {
            _this.scrollTopNotifyed = false;
            _this.scrollBottomNotifyed = false;
        }, this);
    };
    ListView.prototype.isMobile = function () {
        return (cc.sys.isMobile || cc.sys.platform === cc.sys.WECHAT_GAME || cc.sys.platform === cc.sys.QQ_PLAY);
    };
    __decorate([
        property(cc.Prefab)
    ], ListView.prototype, "itemTemplate", void 0);
    __decorate([
        property
    ], ListView.prototype, "spacing", void 0);
    __decorate([
        property
    ], ListView.prototype, "spawnCount", void 0);
    __decorate([
        property(cc.ScrollView)
    ], ListView.prototype, "scrollView", void 0);
    ListView = __decorate([
        ccclass
    ], ListView);
    return ListView;
}(cc.Component));
exports.default = ListView;
// 数据绑定的辅助适配器
var AbsAdapter = /** @class */ (function () {
    function AbsAdapter() {
        this.dataSet = [];
    }
    AbsAdapter.prototype.setDataSet = function (data) {
        this.dataSet = data;
    };
    AbsAdapter.prototype.getCount = function () {
        return this.dataSet.length;
    };
    AbsAdapter.prototype.getItem = function (posIndex) {
        return this.dataSet[posIndex];
    };
    AbsAdapter.prototype._getView = function (item, posIndex) {
        this.updateView(item, posIndex);
        return item;
    };
    return AbsAdapter;
}());
exports.AbsAdapter = AbsAdapter;

cc._RF.pop();
        }
        if (CC_EDITOR) {
            __define(__module.exports, __require, __module);
        }
        else {
            cc.registerModuleFunc(__filename, function () {
                __define(__module.exports, __require, __module);
            });
        }
        })();
        //# sourceMappingURL=ListView.js.map
        